IMPLEMENTATION MODULE OSubs;

(* By: Harrie F.A. de Leeuw *)
(* Part of 'Orbit' program  *)


(************** IMPORT from system libraries *)
FROM MathLib0   IMPORT arctan,
                       sin,
                       cos,
                       sqrt,
                       ln,
                       exp,
                       pi,
                       log;
FROM AESForms   IMPORT FormAlert;

(************** IMPORT from ORBIT specific libraries *)
FROM OrbGlobal  IMPORT LichtSnelheid,
                       RadGrad,
                       Pi,
                       TweePi,
                       AardStraal,
                       AfplatConst,
                       AardMas,
                       SidZon,
                       Sid80,
                       M1e,
                       OmloopNr,
                       Anomalie,
                       Fase,
                       ArgPer,
                       Raan,
                       Azimuth,
                       Elevation,
                       Distance,
                       HeightC,
                       WLongitude,
                       Latitude,
                       HalveAs,
                       DFreq;
FROM OFiler     IMPORT SatRec,
                       Station;



VAR   Mat               :ARRAY [1..3],[1..2] OF REAL;  (* Rotatie matrix *)
      
      X,                           (* Geocentrische coord. van satelliet *)
      Y,                           (*  ,,                                *)
      Z,                           (*  ,,                                *)
      Sid1Jan,                     (* Siderische tijd op 1-Jan v ijkjaar *)
      SinEigenWest,                (* Sinus v eigen westerlengte         *)
      CosEigenWest,                (* Cosinus v eigen westerlengte       *)
      SinEigenNoord,               (* Sinus v eigen noorderbreedte       *)
      CosEigenNoord,               (* Cosinus v eigen noorderbreedte     *)
      EigenX,                      (* Geocentrische coord. eigen station *)
      EigenY,                      (*  ,,                                *)
      EigenZ,                      (*  ,,                                *)
      EccParam1,                   (* Eccentriciteitsparameter           *)
      GravityParam,                (* Invloed ongelijkmatige aantrekking *)
      SinInclinatie,               (* Sinus v inclinatie                 *)
      CosInclinatie,               (* Cosinus v inclinatie               *)
      StartFase,                   (* Fase op ijkpunt                    *)
      RelSpeed,                    (* Snelheid tov eigen station         *)
      PreviousDistance  :REAL;     (* Vorige afstand satelliet-station   *)



PROCEDURE Divide(A,B:REAL):REAL;
VAR Dum: INTEGER;
BEGIN
  IF A=0.0 THEN
    RETURN 0.0
  ELSE
    IF ((B=0.0) OR ((log(ABS(A))-log(ABS(B)))>37.0)) THEN
      Dum:=FormAlert(1,'[1][Divide results in an overflow.| |1E38 is substituted instead!| ][Press RETURN]');
      IF (((A>0.0) & (B>=0.0)) OR ((A<0.0) & (B<0.0))) THEN
        RETURN 1.0E38
      ELSE
        RETURN -1.0E38
      END
    ELSE
      RETURN A/B
    END
  END
END Divide;
      


PROCEDURE Sin(X:REAL):REAL;   (* Goedknoeien sin functie *)
BEGIN
  WHILE X > pi DO
    X:=X - 2.0*pi
  END;
  WHILE X < (1.0E-7 - pi) DO  (* +1.0E-7 X van -pi weg te schuiven *)
    X:=X + 2.0*pi
  END;
  IF X > (pi - 1.0E-7) THEN
    RETURN 0.0
  ELSE
    RETURN sin(X)
  END
END Sin;


PROCEDURE Cos(X:REAL):REAL;   (* Goedknoeien cos functie *)
BEGIN
  WHILE X > (pi/2.0) DO
    X:=X - 2.0*pi
  END;
  WHILE X < (1.0E-7 - 1.5*pi) DO
    X:=X + 2.0*pi
  END;
  IF X > (0.5*pi - 1.0E-7) THEN
    RETURN 0.0
  ELSE
    RETURN cos(X)
  END
END Cos;



PROCEDURE EigenGeocentrisch();
  (* 
     Berekent SinEigenWest, CosEigenWest, SinEigenNoord, CosEigenNoord,
              EigenX, EigenY en EigenZ
     Gebruikt Station.EigenNoord, Station.EigenWest en Station.EigenHoogte
  *)
VAR L8,WN,R9,RT:REAL;
BEGIN
  L8:=Station.EigenNoord*RadGrad;
  SinEigenNoord:=Sin(L8);
  CosEigenNoord:=Cos(L8);
  WN:=Station.EigenWest*RadGrad;
  SinEigenWest:=Sin(0.0-WN);
  CosEigenWest:=Cos(WN);
  R9:=AardStraal*(1.0-AfplatConst/2.0+
        Divide(AfplatConst,2.0*Cos(2.0*L8)))+
        Station.EigenHoogte/1000.0;
  L8:=arctan((1.0-AfplatConst)*(1.0-AfplatConst)*
        Divide(SinEigenNoord,CosEigenNoord));
  RT:=R9*Cos(L8);
  EigenX:=RT*CosEigenWest;
  EigenY:=RT*SinEigenWest;
  EigenZ:=R9*Sin(L8)
END EigenGeocentrisch;


PROCEDURE SiderTijd1Jan();
  (*
     Berekent Sid1Jan
     Gebruikt SatRec.YJaar
  *)
VAR Verschil,Schrikkeldagen:CARDINAL;
BEGIN
  Verschil:=SatRec.YJaar-80;
  IF Verschil=0 THEN
    Schrikkeldagen:=0
  ELSE
    Schrikkeldagen:=TRUNC(FLOAT(Verschil-1)/4.0)
  END;
  Sid1Jan:=(FLOAT(Verschil)*365.0+1.0
                                +FLOAT(Schrikkeldagen))*(SidZon-1.0)+Sid80
END SiderTijd1Jan;


PROCEDURE RelStartTijd(Jaar,Maand,Dag,Uren,Minuten:CARDINAL):REAL;
  (*
     Berekent tijdstip waarop gestart wordt,
     in aantal dagen gerekend vanaf 1 jan 0:00 uur 
     ( =1 !!!!) van het ijkjaar (SatRec.YJaar)
     Gebruikt SatRec.YJaar
  *)
VAR AantalDagen, Ykjaar:CARDINAL;
    Ret,RelSTijd:REAL;
BEGIN
  Ykjaar:=SatRec.YJaar;
  IF Jaar<Ykjaar THEN
    Ret:=0.0-1.0              (* Error *)
  ELSE
    AantalDagen:=0;
    WHILE Jaar#Ykjaar DO
      IF (Ykjaar MOD 4)=0 THEN
        AantalDagen:=AantalDagen+1
      END;
      Ykjaar:=Ykjaar+1;
      AantalDagen:=AantalDagen+365
    END;
    IF (  ( (Ykjaar MOD 4)=0 ) AND ( Maand>2 )   )  THEN
      AantalDagen:=AantalDagen+1
    END;
    Ret:=(FLOAT(AantalDagen+M1e[Maand]+Dag)
                                  +FLOAT(Uren)/24.0+FLOAT(Minuten)/1440.0)
  END;
  RETURN Ret
END RelStartTijd;
  
  
PROCEDURE Macht(Grond,Expo:REAL):REAL;
  (*
     Berekent Grond^Expo
  *)
BEGIN
  RETURN exp(Expo*ln(Grond))
END Macht;


PROCEDURE InitOrbElements();
  (*
     Berekent HalveAs, EccParam1, StartFase, GravityParam,
              SinInclinatie, CosInclinatie
     Gebruikt SatRec
  *)
VAR EccPar,InclRad:REAL;
BEGIN
  WITH SatRec DO
    HalveAs:=
        Macht(Divide(AardMas,(YMeanMotion*YMeanMotion)),0.3333333333333);
    EccPar:=1.0-YEccentr*YEccentr;
    EccParam1:=sqrt(EccPar);
    StartFase:=YMeanAnom/360.0+FLOAT(YOmloop);
    GravityParam:=
        9.95*Divide(Macht(AardStraal/HalveAs,3.5),(EccPar*EccPar));
    InclRad:=YInclinatie*RadGrad;
    SinInclinatie:=Sin(InclRad);
    CosInclinatie:=Cos(InclRad);
  END;
END InitOrbElements;


PROCEDURE BepaalRotMat(T:REAL);
  (*
     Berekent ArgPer, Raan en Mat op tijdstip T
     Gebruikt SatRec, GravityParam, SinInclinatie en CosInclinatie
  *)
VAR RelTijd, RaanRad, ArgPerRad, S0, C0, S2, C2: REAL;
BEGIN
  RelTijd:=T-SatRec.YDag;
  Raan:=SatRec.YRaan-(RelTijd*GravityParam*CosInclinatie);
  RaanRad:=Raan*RadGrad;
  S0:=Sin(RaanRad);
  C0:=Cos(RaanRad);
  ArgPer:=SatRec.YArgPer+RelTijd*GravityParam
                                  *(2.5*CosInclinatie*CosInclinatie-0.5);
  ArgPerRad:=ArgPer*RadGrad;
  S2:=Sin(ArgPerRad);
  C2:=Cos(ArgPerRad);
  Mat[1,1]:=C2*C0-S2*S0*CosInclinatie;
  Mat[1,2]:=0.0-S2*C0-C2*S0*CosInclinatie;
  Mat[2,1]:=C2*S0+S2*C0*CosInclinatie;
  Mat[2,2]:=0.0-S2*S0+C2*C0*CosInclinatie;
  Mat[3,1]:=S2*SinInclinatie;
  Mat[3,2]:=C2*SinInclinatie
END BepaalRotMat;


PROCEDURE BepaalAnomalieOmloop(T:REAL);
  (*
     Berekent Anomalie, OmloopNr en Fase op tijdstip T
     Gebruikt SatRec en StartFase
  *)
VAR RelTijd, Oml, OmMOD: REAL;
    OmlNrMOD: CARDINAL;
BEGIN
  RelTijd:=T-SatRec.YDag;
  Oml:=(SatRec.YAccMeanM*RelTijd+SatRec.YMeanMotion)*RelTijd+StartFase;
  IF Oml>=32767.0 THEN 
    OmMOD:=Oml-32767.0;
    OmlNrMOD:=32767
  ELSE
    OmMOD:=Oml;
    OmlNrMOD:=0
  END;
  OmloopNr:=TRUNC(OmMOD);
  Fase:=TRUNC((OmMOD-FLOAT(OmloopNr))*256.0+0.5);
  Anomalie:=(Oml-FLOAT(OmlNrMOD)-FLOAT(OmloopNr))*TweePi;
  OmloopNr:=OmloopNr+OmlNrMOD
END BepaalAnomalieOmloop;


PROCEDURE SolveKeppler(Time:REAL);
  (*
     Berekent X, Y, Z en HeightC op tijdstip Time
     Gebruikt Anomalie, SatRec, Mat, Sid1Jan, EccParam1 en HalveAs
  *)
VAR Ready: BOOLEAN;
    E, S3, C3, R3, M5, X0, Y0, X1, Y1, G7, S7, C7: REAL;
    Count: CARDINAL;
    Dum: INTEGER;
BEGIN
  Count:=0;
  E:=Anomalie+SatRec.YEccentr*Sin(Anomalie)
               +0.5*SatRec.YEccentr*SatRec.YEccentr*Sin(Anomalie+Anomalie);
  Ready:=FALSE;
  REPEAT
    Count:=Count+1;
    S3:=Sin(E);
    C3:=Cos(E);
    R3:=1.0-SatRec.YEccentr*C3;
    M5:=E-SatRec.YEccentr*S3-Anomalie;
    IF Count>100 THEN
      Dum:=FormAlert(1,'[1][I am having trouble solving|the Keppler equations.| |Results may be wrong!][Press RETURN]');
      Ready:=TRUE
    END;
    IF (ABS(M5) < 1.0E-6) THEN
      Ready:=TRUE
    ELSE
      E:=E-Divide(M5,R3)
    END
  UNTIL Ready;
  X0:=HalveAs*(C3-SatRec.YEccentr);
  Y0:=HalveAs*EccParam1*S3;
  HeightC:=HalveAs*R3;
  X1:=X0*Mat[1,1]+Y0*Mat[1,2];
  Y1:=X0*Mat[2,1]+Y0*Mat[2,2];
  Z :=X0*Mat[3,1]+Y0*Mat[3,2];
  G7:=Time*SidZon+Sid1Jan;
  G7:=(G7-FLOAT(TRUNC(G7)))*TweePi;
  S7:=0.0-Sin(G7);
  C7:=Cos(G7);
  X :=X1*C7-Y1*S7;
  Y :=X1*S7+Y1*C7;
END SolveKeppler;



PROCEDURE SolveParams(Time, DeltaTime:REAL);
  (*
     Berekent Distance, RelSpeed, Elevation, Azimuth, WLongitude en
              Latitude op tijdstip Time
              DeltaTime is de tijdstap in seconden (nodig voor RelSpeed)
     Gebruikt HeightC, X, Y, Z, EigenX, EigenY, EigenZ,
              SinEigenWest, CosEigenWest, SinEigenNoord en CosEigenNoord
  *)

VAR DeltaDistance, X5, Y5, Z5, X8, Y8, Z8, S5, C5, B5: REAL;

  PROCEDURE ArcTang(Xc,Yc:REAL):REAL;
    (*
       Berekent Arctangens in juiste quadrant
    *)
  VAR AT:REAL;
  BEGIN
    IF Xc # 0.0 THEN
      AT:=arctan(Divide(Yc,Xc));
      IF AT < 0.0 THEN AT:=AT+Pi END;
      IF Yc < 0.0 THEN AT:=AT+Pi END;
      RETURN AT
    ELSE
      IF Yc >= 0.0 THEN
        RETURN (Pi/2.0)
      ELSE
        RETURN (1.5*Pi)
      END
    END
  END ArcTang;

BEGIN
  X5:=X-EigenX;
  Y5:=Y-EigenY;
  Z5:=Z-EigenZ;
  Distance:=sqrt(X5*X5+Y5*Y5+Z5*Z5);
  DeltaDistance:=Distance-PreviousDistance;
  RelSpeed:=Divide(DeltaDistance,DeltaTime);
  PreviousDistance:=Distance;
  X8:=0.0-X5*CosEigenWest*SinEigenNoord-Y5*SinEigenWest*SinEigenNoord
                                                         +Z5*CosEigenNoord;
  Y8:=Y5*CosEigenWest-X5*SinEigenWest;
  Z8:=X5*CosEigenWest*CosEigenNoord+Y5*SinEigenWest*CosEigenNoord
                                                         +Z5*SinEigenNoord;
  S5:=Divide(Z8,Distance);
  C5:=sqrt(1.0-S5*S5);
  Elevation:=arctan(Divide(S5,C5))/RadGrad;
  Azimuth:=ArcTang(X8,Y8)/RadGrad;
  WLongitude:=360.0-ArcTang(X,Y)/RadGrad;
  B5:=Divide(Z,HeightC);
  Latitude:=arctan(Divide(B5,sqrt(1.0-B5*B5)))/RadGrad
END SolveParams;



PROCEDURE SolveDoppler;
  (*
     Berekent DFreq
     Gebruikt SatRec.Frekwentie en RelSpeed
  *)
BEGIN
  DFreq:=((Divide((0.0-SatRec.Frekwentie*1.0E6),LichtSnelheid))*RelSpeed)
END SolveDoppler;


BEGIN  (* OrbSubs initialisatiecode *)
  PreviousDistance:=0.0
END OSubs.
